# coding: utf8
__doc__ = 'instrument class with basic port communication'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
import time
import logging
from . import Env, Utils

if Env.LOGGING_FORMAT:
    logging.BASIC_FORMAT = Env.LOGGING_FORMAT
else:
    logging.BASIC_FORMAT = '%(asctime)s ' + logging.BASIC_FORMAT

class PortNotFound( Exception ):
    pass

class UnknownPortError( Exception ):
    pass

class ResponseError( Exception ):
    pass

class CommandSyntaxError( Exception ):
    pass

class CommandSemanticsError( Exception ):
    pass

class CommandExecuteError( Exception ):
    pass

class CommandTimeoutError( Exception ):
    pass

class IDNMatchError( Exception ):
    pass




class Instrument:
    '''abstract instrument class'''
    
    DEFAULT_NAME = 'INST'
    DEFAULT_TERMINATOR_WRITE = '\x0A'  # '\n'
    DEFAULT_TERMINATOR_READ = '\x0A'  # '\n'
    DEFAULT_TERMINATOR_RESET = '\x03'  # Ctrl-C
    DEFAULT_TIMEOUT = 5
    DEFAULT_PROMPTS = re_compile( '[=#?!]>' )
    DEFAULT_PROMPTS_MULTILINE = re_compile( '[=#?!]?>' )
    DEFAULT_IDN = None
    DEFAULT_REBOOT_RETRY = 10
    DEFAULT_LINE_LIMIT = 128
    DEFAULT_CHECK_RETURN_COMMAND = True
   

    def __init__( self, *args, **kwargs ):
        '''init'''
        # logging level 
        self.verbose = Env.VERBOSE
        self.debug = Env.DEBUG
        self.info = Env.INFO
        self.warning = Env.WARNING
        if self.debug:
            level = logging.DEBUG
        elif self.info:
            level = logging.INFO
        elif self.warning:
            level = logging.WARNING
        else:
            level = logging.FATAL
        logging.basicConfig( level=level, format=Env.LOG_FORMAT, datefmt=Env.LOG_DATEFMT )
        self.logger = logging.getLogger( self.DEFAULT_NAME )
        self.check_return_command = self.DEFAULT_CHECK_RETURN_COMMAND
        self.returned_cmd = None
        self.returned_prompt = None

        # load from parms/env/default and saved as attributes
        self.logger.debug( '__init__: args: %s'% str(args) )
        try:
            kwargs['port'] = args[0]
        except:
            pass
        if not 'port' in kwargs:
            kwargs['port'] = Env.PORT
        if not 'baudrate' in kwargs:
            kwargs['baudrate'] = Env.BAUDRATE
        if not 'rtscts' in kwargs:
            kwargs['rtscts'] = Env.RTSCTS
        if not 'parity' in kwargs:
            kwargs['parity'] = Env.PARITY
        if not 'connect' in kwargs:
            kwargs['connect'] = True
        if not 'prompts' in kwargs:
            kwargs['prompts'] = self.DEFAULT_PROMPTS
        if not 'timeout' in kwargs:
            kwargs['timeout'] = self.DEFAULT_TIMEOUT
        if not 'check_idn' in kwargs:
            kwargs['check_idn'] = True
        if not 'terminal_reset' in kwargs:
            kwargs['terminal_reset'] = True
        # some attributes 'connect', ...  need to be renamed for method conflict
        for n in ['connect', 'timeout']:
            kwargs['_'+n] = kwargs.pop(n)
        # attached as attributes
        for k, v in kwargs.items():
            self.__dict__[k] = v
        self.logger.debug( '__init__: %s'% str(self.__dict__) )

        # init/connect
        self.idn = None
        self.serial_number = None
        self.port = self.PORT_TYPE(self, *args, **kwargs)
        if self._connect:
            self.connect()

    @property        
    def connected( self ):
        return self.port.connected

    def setVerbose( self, verbose ):
        '''Set verbose mode'''
        self.verbose = verbose

    def setLoggingLevel( self, level ):
        '''set logging level'''
        self.debug, self.info = False, False
        if level in ['debug']:
            self.debug = True 
            logging.basicConfig( level=logging.DEBUG )
        elif level in ['info']:
            self.info = True 
            logging.basicConfig( level=logging.INFO )
        else:
            logging.basicConfig( level=logging.FATAL )
 
    def setInfo( self, debug ):
        '''set debug'''
        self.debug = debug
 
    def setPrompts( self, prompts=None ):
        '''dynamically modify the prompts'''
        old_prompts = self.prompts
        if prompts is None:
            prompts = self.DEFAULT_PROMPTS
        self.prompts = prompts
        return old_prompts

    def setTimeout( self, new=None ):
        if new is None:
            new = self.DEFAULT_TIMEOUT
        old = self.port.timeout
        self.port.timeout = new
        return old

    def connect( self, check_idn=True ):
        '''connect'''
        self.port.connect()
        if not self.port.connected:
            return
        if self.terminal_reset and self.DEFAULT_TERMINATOR_RESET:
            self.port.write( self.DEFAULT_TERMINATOR_RESET )
            self.port.flush()
            self.readUntilPrompts()
        if check_idn and self.check_idn and self.DEFAULT_IDN is not None:
            self.scpiIdn()

    def disconnect( self ):
        '''disconnect'''
        if self.port.connected:
            self.port.disconnect()

    def assertIsOpen( self ):
        '''assert port is opened'''
        if not self.port.connected:
            self.port.connect()
            if not self.port.connect:
                raise Exception("Fail to open port")
  
    def readUntilPrompts( self, line_callback=None ):
        '''read until prompts'''
        contents, newline_lst, newline_str = [], [], ''
        while True:
            byte = self.port.read(1)
            if byte:
                if Env.PYTHON_V3:
                    byte = chr(ord(byte))
                if byte == self.DEFAULT_TERMINATOR_READ:
                    newline_str = newline_str.rstrip()
                    contents.append( newline_str )
                    #self.logger.debug( '[r] '+ newline_str )
                    if line_callback is not None:
                        # use this carefully
                        line_callback( newline_str )
                    newline_lst, newline_str = [], ''
                else:
                    newline_lst.append( byte )
                    newline_str += byte
            else:
                contents.append( newline_str )
                if contents:
                    raise CommandTimeoutError( ' | '.join(contents) )
                else:
                    raise CommandTimeoutError( 'No response' )
            #print( newline_str )  # for port debug
            match = self.prompts.match( newline_str )
            if match:
                contents.append( newline_str )
                #self.logger.debug( '[P] '+ newline_str )
                return contents

    def readLine( self, eol='\n', timeout=None ):
        chars = []
        t0 = time.time()
        while True:
            char = self.port.read(1) 
            if char:
                if char == eol:
                    break
                chars.append( char )
            elif timeout:
                if time.time() > t0 + timeout:
                    break
        return ''.join(chars).rstrip()

    def writeLine( self, dat ):
        self.assertIsOpen() 
        #print(type(dat), dat)
        if Env.PYTHON_V3:
            if isinstance( dat, str ):
                dat = dat.encode('utf8')
        else:
            if isinstance( dat, unicode ):
                dat = dat.encode('utf8')
        self.port.write( dat )
        self.port.write( self.DEFAULT_TERMINATOR_WRITE )
        self.port.flush()
   
    def writeCommand( self, cmd ):
        '''write command and wait for prompts'''
        cmd = cmd.strip()
        self.writeLine( cmd )
        self.logger.debug( '[T] '+cmd )
        ret = self.readUntilPrompts()
        for line in [i.strip() for i in ret]:
            self.logger.debug( '[R] '+ line )
        self.checkReturnedPrompt( ret )
        if self.check_return_command:
            self.checkReturnedCommand( ret, cmd )
        return ret[1:-1] 
    
    def writeCommandRetry( self, cmd, retry=None ):
        '''write command with retry '''
        if retry is None:
            retry = Env.COMMAND_FAIL_RETRY
        assert retry > 1
        for r in range(retry-1):
            try:
                ret = self.writeCommand( cmd )
                return ret
            except Exception as e:
                if Env.VERBOSE:
                    print( e )
        return self.writeCommand( cmd )
  
    def checkReturnedCommand( self, ret, cmd ):
        '''assert command returned is valid'''
        if Env.NO_ECHO_CHECK:
            return
        cmdret = ret[0]
        if not cmd:
            return
        if Env.PYTHON_V3 and isinstance(cmd, bytes):
            cmdret = cmdret.encode('utf8')
        if cmd != cmdret:
            raise ResponseError('Command %s, but returned %s'% (cmd, cmdret))

    def checkReturnedPrompt( self, ret ):
        '''assert prompt is valid'''
        self.returned_cmd = cmd = ret[0].strip()
        self.returned_prompt = prompt = ret[-1].strip()
        if prompt in ['?>', '?']:
            result = ret[1:-1]
            if result:
                raise CommandSyntaxError( cmd + ', returns: ' + ','.join(result) )
            else:
                raise CommandSyntaxError( cmd )
        elif prompt == '!>':
            result = ret[1:-1]
            if result:
                raise CommandExecuteError( cmd + ', returns: ' + ','.join(result) )
            else:
                raise CommandExecuteError( cmd )

    # Instrument class only supports basic commands:
    # 
    # 1. scpi identify
    #    =>*idn?
    #    mcush,1.1          --- model, version
    #    NNNNNNNNNNNNNNNN   --- serial number (if exists)
    # 
    # 2. scpi reset
    #    =>*rst
    # 
    # 3. reboot cpu core
    #    =>reboot
    # 

    def scpiRst( self ):
        '''scpi reset'''
        self.writeCommand( '*rst' )

    def scpiIdn( self, check=True ):
        '''get identify name'''
        ret = self.writeCommand( '*idn?' )
        self.idn = ret[0].strip()
        if len(ret)>1:
            self.serial_number = ret[1].strip()
        self.logger.info( 'IDN:%s', str(self.idn) )
        if check and (not Env.NO_IDN_CHECK):
            if not self.DEFAULT_IDN.match( self.idn ):
                raise IDNMatchError(self.idn.split(',')[0])
        return self.idn

    def reboot( self, delay=None ):
        '''reboot command'''
        sync = False
        retry = 0
        try:
            self.writeCommand( 'reboot' )
            self.connect()
            sync = True
        except ResponseError:
            pass
        except CommandTimeoutError:
            pass
        while not sync: 
            try:
                SerialInstrument.connect( self )
                sync = True
            except:
                retry = retry + 1
                if retry > self.DEFAULT_REBOOT_RETRY:
                    raise CommandTimeoutError()
        self.scpiIdn()
        if delay is None:
            time.sleep( Env.DELAY_AFTER_REBOOT )
        else:
            time.sleep( delay )

    def getRebootTimes( self ):
        cmd = 'reboot -c'
        self.writeCommand( cmd )

    def getModel( self ):
        if self.idn is None:
            self.scpiIdn()
        try:
            return self.idn.split(',')[0]
        except IndexError:
            return ''

    def getVersion( self ):
        if self.idn is None:
            self.scpiIdn()
        try:
            return self.idn.split(',')[1]
        except IndexError:
            return ''

    def getSerialNumber( self ):
        if self.serial_number is None:
            ret = self.writeCommand( '*idn?' )
            if len(ret) > 1:
                self.serial_number = ret[1].strip()
        self.logger.info( 'SN:%s', str(self.serial_number) )
        if self.serial_number is None:
            return ''
        return self.serial_number

    def getIntegerSerialNumber( self, msb=True ):
        sn = [Utils.s2B(b) for b in Utils.unhexlify(self.getSerialNumber())]
        ret = 0 
        while sn:
            if msb:
                ret = (ret<<8) | sn.pop(0)
            else:
                ret = (ret<<8) | sn.pop()
        return ret
 
    def printInfo( self ):
        print( '%s, %s'% (self.getModel(), self.getVersion()) )

    # NOTE: reboot counter is not supported in some platform   
    def getRebootCounter( self ):
        cmd = 'reboot -c'
        ret = self.writeCommand( cmd )
        return int(ret[0])

    def resetRebootCounter( self ):
        cmd = 'reboot -r'
        self.writeCommand( cmd )

    def checkCommand( self, name ):
        cmd = '? -c %s'% name
        return bool(int(self.writeCommand(cmd)[0])) 


class Port(object):
    
    def __init__( self, parent, *args, **kwargs ):
        self.parent = parent
        for k, v in kwargs.items():
            self.__dict__[k] = v
        self._connected = False

    def __del__( self ):
        self.disconnect()
 
    def connect( self ):
        raise NotImplementedError

    def disconnect( self ):
        pass
    
    @property        
    def connected( self ):
        return self._connected

    def read( self, read_bytes=1 ):
        raise NotImplementedError

    def write( self, buf ):
        raise NotImplementedError
 
    def flush( self ):
        pass

    def readall( self ):
        raise

    def update_timeout( self, timeout ):
        pass 

    @property
    def timeout( self ):
        return self._timeout

    @timeout.setter
    def timeout( self, val ):
        self._timeout = val
        self.update_timeout( val )
 
 
class SerialPort(Port):
    
    def __init__( self, parent, *args, **kwargs ):
        import serial
        self.serial_exception = serial.SerialException 
        try:
            self.ser = serial.serial_for_url( self.port, do_not_open=True )
        except AttributeError:
            self.ser = serial.Serial()
        Port.__init__( self, parent, *args, **kwargs )
   
    def connect( self ):
        if self._connected:
            return
        self.ser.port = self.port
        self.ser.baudrate = self.baudrate
        self.ser.rtscts = self.rtscts
        self.ser.parity = self.parity
        self.ser.timeout = self.timeout
        try:
            self.ser.open()
            self._connected = True
        except IOError:
            raise PortNotFound( self.port )
        except Exception as e:
            #print( e )
            raise UnknownPortError( e )
        if self._connected:
            try:
                self.ser.reset_input_buffer()
                self.ser.reset_output_buffer()
            except:
                pass

    def disconnect( self ):
        if self._connected:
            self.ser.close()
            self._connected = False
 
    def update_timeout( self, timeout ):
        self.ser.timeout = timeout
        
    def read( self, read_bytes=1 ):
        try:
            return self.ser.read(read_bytes)
        except self.serial_exception as e:
            raise UnknownPortError( str(e) )
    
    def readall( self ):
        return self.read( self.ser.in_waiting )

    def write( self, buf ):
        try:
            #print(type(buf), buf)
            if Env.PYTHON_V3 and isinstance(buf, str):
                buf = buf.encode("utf8")
            self.ser.write( buf )
        except self.serial_exception as e:
            raise UnknownPortError( str(e) )
 
    def flush( self ):
        try:
            self.ser.flush()
        except self.serial_exception as e:
            raise UnknownPortError( str(e) )
 

class SocketPort(Port):
        
    def __init__( self, parent, *args, **kwargs ):
        import socket
        self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        Port.__init__( self, parent, *args, **kwargs )
        
    def connect( self ):
        if self._connected:
            return
        self.s.connect( (self.ip, self.port) )
        self.s.settimeout( self.timeout )
        self._connected = True

    def disconnect( self ):
        self._connected = False
        self.s.close()
    
    def read( self, read_bytes=1 ):
        if self._connected:
            ret = []
            for i in range(read_bytes):
                r = self.s.recv(1)
                if r:
                    ret.append( r )
                else:
                    break
            #print( 'read', ret )
            return ''.join(ret)

    def write( self, buf ):
        if self._connected:
            self.s.sendall(buf)
 




class SerialInstrument(Instrument):
    '''Serial port based instruments'''
    PORT_TYPE = SerialPort


class SocketInstrument(Instrument):
    '''Socket port based instruments'''
    PORT_TYPE = SocketPort





# coding:utf8
__doc__ = 'mcu registers'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from . import Utils


REGISTER_TYPE = {
# c_type_name: (bit_width, bytes, get_func, set_func)
'int8':   ( 8,  1, Utils.s2b, Utils.b2s ),
'uint8':  ( 8,  1, Utils.s2B, Utils.B2s ),
'int16':  ( 16, 2, Utils.s2h, Utils.h2s ),
'uint16': ( 16, 2, Utils.s2H, Utils.H2s ),
'int32':  ( 32, 4, Utils.s2i, Utils.i2s ),
'uint32': ( 32, 4, Utils.s2I, Utils.I2s ),
'int64':  ( 64, 8, Utils.s2q, Utils.q2s ),
'uint64': ( 64, 8, Utils.s2Q, Utils.Q2s ),
'float':  ( 32, 4, Utils.s2f, Utils.f2s ),
'double': ( 64, 8, Utils.s2d, Utils.q2s ),
}

class Register:
    '''mcu register'''
    REG_TYPE = None

    def __init__( self, name, address, description, reg_type=None ):
        self.name = str(name)
        self.address = int(address)
        self.description = str(description)
        self.reg_type = str(reg_type) if reg_type is not None else self.REG_TYPE
        assert( self.reg_type in REGISTER_TYPE )

    def __str__( self ):
        return self.name

class Int8(Register):
    REG_TYPE = 'int8'

class UInt8(Register):
    REG_TYPE = 'uint8'

class Int16(Register):
    REG_TYPE = 'int16'

class UInt16(Register):
    REG_TYPE = 'uint16'

class Int32(Register):
    REG_TYPE = 'int32'

class UInt32(Register):
    REG_TYPE = 'uint32'

class Int64(Register):
    REG_TYPE = 'int64'

class UInt64(Register):
    REG_TYPE = 'uint64'

class Float(Register):
    REG_TYPE = 'float'

class Double(Register):
    REG_TYPE = 'double'





# coding: utf8
__doc__ = 'INA219 Current/Power Monitor'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils


REG_CONFIG  = 0
REG_SHUNT   = 1
REG_BUS     = 2
REG_POWER   = 3
REG_CURRENT = 4
REG_CALIB   = 5

ADC_9B      = 0
ADC_10B     = 1
ADC_11B     = 2
ADC_12B     = 8
ADC_12B_2   = 9
ADC_12B_4   = 10
ADC_12B_8   = 11
ADC_12B_16  = 12
ADC_12B_32  = 13
ADC_12B_64  = 14
ADC_12B_128 = 15

BRNG_16 = 0x0000
BRNG_32 = 0x2000

PG_1 = (0<<11) 
PG_2 = (1<<11) 
PG_4 = (2<<11) 
PG_8 = (3<<11) 

MODE_POWER_DOWN     = 0
MODE_SHUNT_TRIG     = 1
MODE_BUS_TRIG       = 2
MODE_SHUNT_BUS_TRIG = 3
MODE_ADC_OFF        = 4
MODE_SHUNT_CONT     = 5
MODE_BUS_CONT       = 6
MODE_SHUNT_BUS_CONT = 7



class Ina219:
 
    mode = MODE_SHUNT_BUS_CONT
    badc = ADC_12B_128  # 12 bit, slow mode
    sadc = ADC_12B_128  # 12 bit, slow mode
   
    rshunt = 0.1  # in ohms
   
    # ref: https://github.com/adafruit/Adafruit_INA219/blob/master/Adafruit_INA219.cpp
    # Adafruit_INA219::setCalibration_16V_400mA 
    brng = BRNG_16
    gain = PG_1
    vshunt_max = 0.32
    current_lsb = 1/20000.0
    power_lsb = 0.001
    volt_lsb = 0.001
    calib = 8192
     
    def __init__( self, controller, scl=None, sda=None, addr=0 ):
        self.controller = controller
        self.controller.i2c_init( 0x40+(addr&0xF), scl=scl, sda=sda )
        self.current_reg = None
        self.config()
       
    def writeReg( self, reg, dat ):
        self.controller.i2c( [reg, dat>>8, dat&0xFF] )
        self.current_reg = reg
  
    def readReg( self, reg ):
        if self.current_reg != reg:
            r = self.controller.i2c( [reg], read=2 )
            self.current_reg = reg
        else:
            r = self.controller.i2c( read=2 )
        return (r[0]<<8)+r[1]

    def config( self ):
        cfg = self.brng + self.gain + (self.badc<<7) + (self.sadc<<3) + self.mode
        self.writeReg( REG_CONFIG, cfg )
        self.writeReg( REG_CALIB, self.calib )

    def readAllRegs( self ):
        ret = {}
        ret['config'] = self.readReg( REG_CONFIG )
        ret['shunt'] = self.readReg( REG_SHUNT )
        ret['bus'] = self.readReg( REG_BUS )
        ret['power'] = self.readReg( REG_POWER )
        ret['current'] = self.readReg( REG_CURRENT )
        ret['calib'] = self.readReg( REG_CALIB )
        return ret

    def readVoltage( self ):
        v = self.readReg( REG_BUS )
        if v & 1:
            raise 'volt overflow'
        return (v>>1)*self.volt_lsb

    def readCurrent( self ):
        v = Utils.H2h(self.readReg( REG_CURRENT ))
        return v*self.current_lsb

    def readPower( self ):
        v = self.readReg( REG_POWER )
        return v*self.power_lsb




# coding: utf8
__doc__ = 'products from Zhenzhou Feeltech'
__author__ = 'Peng Shulin <trees_peng@163.com>'
import re
import time
from .. import Env, Instrument

MODES_A = [
'sine',
'square',  # fixed 50% duty
'rectangle',  # adjustable duty
'trapezoid',
'cmos',
'adj-pulse',
'dc',
'triangle',
'ramp',
'neg ramp',
'stair triangle',
'stair step',
'neg stair step',
'pos exponent',
'neg exponent',
'pos fall exponent',
'neg fall exponent',
'pos logarithm',
'neg logarithm',
'pos fall logarithm',
'neg fall logarithm',
'pos full sine',
'neg full sine',
'pos half sine',
'neg half sine',
'lorentz pulse',
'multitone',
'random noise',
'ecg',
'trapezoid',
'sinc pulse',
'impulse',
'awgn',
'am',
'pm',
'chirp',
]

MODES_B = [
'sine',
'square',
'rectangle',
'trapezoid',
'cmos',
'dc',
'triangle',
'ramp',
'neg ramp',
'stair triangle',
'stair step',
'neg stair step',
'pos exponent',
'neg exponent',
'pos fall exponent',
'neg fall exponent',
'pos logarithm',
'neg logarithm',
'pos fall logarithm',
'neg fall logarithm',
'pos full sine',
'neg full sine',
'pos half sine',
'neg half sine',
'lorentz pulse',
'multitone',
'random noise',
'ecg',
'trapezoid',
'sinc pulse',
'impulse',
'awgn',
'am',
'pm',
'chirp',
]

for i in range(64):
    MODES_A.append('arb%d'%(i+1))
    MODES_B.append('arb%d'%(i+1))

KEYS = ['wave', 'meas', 'sweep', 'mod', 'sync', 'sys', 'more', 
'ch1', 'ch2', 'f1', 'f2', 'f3', 'f4', 'f5',
'left', 'right', 'ok', 'up', 'down']

MEASURE_GATE = [1, 10, 100]

SWEEP_MODE = ['freq', 'amp', 'offset', 'duty']


class FY6900( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'FY6900'
    DEFAULT_TERMINATOR_READ = '\x0A'
    DEFAULT_TERMINATOR_WRITE = '\x0A'
    DEFAULT_TERMINATOR_RESET = 'UMO\x0A'
    DEFAULT_CHECK_RETURN_COMMAND = False

    def __init__( self, *args, **kwargs ):
        kwargs['baudrate'] = 115200  # this is fixed
        kwargs['check_idn'] = False  # not support scpi commands
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 

    def connect( self, check_idn=False ):
        Instrument.SerialInstrument.connect( self, check_idn=False )
        self.readUntilPrompts()
        self.readUntilPrompts()
        self.readUntilPrompts()
        # read model/id
        model = self.getModel()
        sn = self.getSerialNumber()
        self.idn = ','.join([model,sn])

    def readUntilPrompts( self, line_callback=None ):
        # this device response at most one line, but followed with multiple 0x0A 
        newline_str = ''
        while True:
            byte = self.port.read(1)
            if byte:
                if Env.PYTHON_V3:
                    byte = chr(ord(byte))
                if byte == self.DEFAULT_TERMINATOR_READ:
                    newline_str = newline_str.rstrip()
                    self.logger.debug( newline_str )
                    return [newline_str]
                else:
                    newline_str += byte
            else:
                if newline_str:
                    raise Instrument.CommandTimeoutError( newline_str )
                else:
                    raise Instrument.CommandTimeoutError( 'No response' )

    def writeCommand( self, cmd ):
        # write command and wait for response
        cmd = cmd.strip()
        self.writeLine( cmd )
        self.logger.debug( cmd )
        ret = self.readUntilPrompts()
        for line in [i.strip() for i in ret]:
            if line:
                self.logger.debug( line )
        return ret
 
    def getModel( self ):
        ret = self.writeCommand("UMO")
        self.readUntilPrompts()
        self.readUntilPrompts()
        self.readUntilPrompts()
        return ret[0]

    def getSerialNumber( self ):
        ret = self.writeCommand("UID")
        self.readUntilPrompts()
        return ret[0]
       
    def outputEnableA( self, enable=True ):
        self.writeCommand( 'WMN1' if enable else 'WMN0' )
    
    def outputEnableB( self, enable=True ):
        self.writeCommand( 'WFN1' if enable else 'WFN0' )

    def outputDisableA( self ):
        self.outputEnableA( False )
    
    def outputDisableB( self, enable=True ):
        self.outputEnableB( False )

    def beepEnable( self, enable=True ):
        self.writeCommand( 'UBZ1' if enable else 'UBZ0' )
    
    def beepDisable( self ):
        self.beepEnable( False )

    def beepIsEnabled( self ):
        r = self.writeCommand( 'RBZ' )
        return bool(r > 0)

    def key( self, name ):
        key_id = KEYS.index(name)
        self.writeCommand( 'KEY%d'% key_id )     
    
    def outputA( self, mode, amp, freq, offset=0, duty=None, phase=0, enable=True ):
        if isinstance(mode, int):
            self.writeCommand( 'WMW%02d'% mode )
        else:
            self.writeCommand( 'WMW%02d'% MODES_A.index(mode) )
        self.writeCommand( 'WMA%f'% float(amp) )  # in V
        self.writeCommand( 'WMF%d'% int(freq*1E6) )  # in uHz
        self.writeCommand( 'WMO%f'% float(offset) )  # in V
        if duty is not None:
            self.writeCommand( 'WMD%.1f'% float(duty) )  # in %
        self.writeCommand( 'WMP%f'% float(phase) )  # in degree
        if enable:
            self.outputEnableA( True )

    def outputB( self, mode, amp, freq, offset=0, duty=None, phase=0, enable=True ):
        if isinstance(mode, int):
            self.writeCommand( 'WFW%02d'% mode )
        else:
            self.writeCommand( 'WFW%02d'% MODES_B.index(mode) )
        self.writeCommand( 'WFA%f'% float(amp) )  # in V
        self.writeCommand( 'WFF%d'% int(freq*1E6) )  # in uHz
        self.writeCommand( 'WFO%f'% float(offset) )  # in V
        if duty is not None:
            self.writeCommand( 'WFD%.1f'% float(duty) )  # in %
        self.writeCommand( 'WFP%f'% float(phase) )  # in degree
        if enable:
            self.outputEnableB( True )

    def measureConfig( self, dc=True, gate=1 ):
        self.meas_gate = MEASURE_GATE.index(gate)
        self.meas_dc = dc
        self.writeCommand( 'WCG%d'% self.meas_gate )
        self.writeCommand( 'WCC%d'% (1 if dc else 0) )
        self.writeCommand( 'WCZ' )

    def measureReadFreq( self ):
        freq = float(self.writeCommand( 'RCF' )[0])
        freq /= 10**self.meas_gate
        self.readUntilPrompts()
        return freq
    
    def measureReadCounter( self ):
        counter = float(self.writeCommand( 'RCC' )[0])
        self.readUntilPrompts()
        return counter
        
    def sweepConfig( self, mode, start, end, time, linear=True, auto_start=True ):
        self.sweep_mode = SWEEP_MODE.index(mode)
        self.sweep_start = start
        self.sweep_end = end
        self.sweep_time = time
        self.sweep_linear = linear
        self.writeCommand( 'SBE0' )
        self.writeCommand( 'SOB%d'% self.sweep_mode )
        self.writeCommand( 'SST%s'% self.sweep_start )
        self.writeCommand( 'SEN%s'% self.sweep_end )
        self.writeCommand( 'STI%s'% self.sweep_time )
        self.writeCommand( 'SMO%d'% (0 if linear else 1) )
        if auto_start:
            self.writeCommand( 'SBE1' )

    def sweepStart( self ):
        self.writeCommand( 'SBE1' )

    def sweepStop( self ):
        self.writeCommand( 'SBE0' )




# coding:utf8
__doc__ = 'Fluke 8808A 5-1/2 digit bench-top multimeter'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env


FUNCTION = ['AAC', 'AACDC', 'ADC', 'CONT', 'DIODE', 'FREQ', 'OHMS', 'VAC', 'VACDC', 'VDC' ]
FUNCTION2 = ['AAC', 'ADC', 'FREQ', 'OHMS', 'VAC', 'VDC' ]
RANGE = {
'VAC':   [0.2, 2, 20, 200, 1000],
'VACDC': [0.2, 2, 20, 200, 1000],
'VDC':   [0.2, 2, 20, 200, 1000],
'AAC':   [0.02, 0.2, 2, 10],
'AACDC': [0.02, 0.2, 2, 10],
'ADC':   [0.02, 0.2, 2, 10],
'OHMS':  [200, 2000, 20000, 200000, 2000000, 20000000, 100000000],
'FREQ':  [2000, 20000, 200000, 1000000],
}


class F8808A( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'F8808A'
    DEFAULT_TIMEOUT = 10
    DEFAULT_PROMPTS = re_compile( '(\x03)?[=#?!]>' )
    DEFAULT_IDN = re_compile( 'FLUKE, 8808A, .*' )

    function = None
    function2 = None

    def writeCommand( self, cmd ):
        # write command and wait for prompts
        cmd = cmd.upper()
        self.writeLine( cmd )
        self.logger.debug( cmd )
        ret = self.readUntilPrompts()
        for line in [i.strip() for i in ret]:
            if line:
                self.logger.debug( line )
        if not ret[0]:
            ret = ret[1:]
        self.checkReturnedPrompt( ret )
        self.checkReturnedCommand( ret, cmd )
        return ret[1:-1] 

    def setFormat( self, sel=1 ):
        # format 1: 1.2345E+6
        # format 2: +12.345E+6 OHM
        self.writeCommand( 'format %d'% sel ) 
       
    def getFormat( self ):
        return int(self.writeCommand( 'format?' )[0])
         
    def setRate( self, rate='m' ):
        # rate: S/M/F
        self.writeCommand( 'rate %s'% rate ) 

    def function( self, mode, wire=None, range=None ):
        # set main function
        mode = mode.strip().upper()
        if not mode in FUNCTION:
            raise Exception("invalid mode %s"% mode)
        self.function = mode
        self.writeCommand( self.function )
        if mode not in FUNCTION2:
            self.function2
        if self.getFormat() != 2:
            self.setFormat(2)
        if mode not in ['CONT', 'DIODE']:
            self.setRange(range)
  
    def function2( self, mode, wire=None ):
        # set secondary function
        mode = mode.strip().upper()
        if not mode in FUNCTION2:
            raise Exception("invalid mode %s"% mode)
        self.function2 = mode+'2'
        self.writeCommand( self.function2 )
  
    def setRange( self, value=None ):
        # set range
        if value is None:
            self.writeCommand( 'AUTO' )
        else:
            self.writeCommand( 'FIXED' )
            idx = RANGE[self.function].index(value)+1
            self.writeCommand( 'RANGE %d'% idx )
     
    def getReading( self ):
        # read the first display reading
        r, u = self.writeCommand( "MEAS1?" )[0].split()
        return (float(r), u)

    def getReading2( self ):
        # read the second display reading
        try:
            r, u = self.writeCommand( "MEAS2?" )[0].split()
            return (float(r), u)
        except Instrument.CommandExecuteError:
            return (None, None)
    
    def hold( self, on=True ):
        self.writeCommand( 'HOLD' if on else 'HOLDCLR' )
         



# coding: utf8
__doc__ = 'MAX31865 SPI Thermometer'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils


REG_CONFIG       = 0
REG_RESULT_H     = 1
REG_RESULT_L     = 2
REG_HIGH_FAULT_H = 3
REG_HIGH_FAULT_L = 4
REG_LOW_FAULT_H  = 5
REG_LOW_FAULT_L  = 6
REG_FAULT        = 7

CONFIG_BIAS_ON   = 0x80
CONFIG_BIAS_OFF  = 0x00
CONFIG_CONV_AUTO = 0x40
CONFIG_CONV_OFF  = 0x00
CONFIG_1SHOT_ON  = 0x20
CONFIG_1SHOT_OFF = 0x00
CONFIG_3WIRE_ON  = 0x10
CONFIG_3WIRE_OFF = 0x00
CONFIG_FILTER_50 = 0x01
CONFIG_FILTER_60 = 0x00
CONFIG_FAULT_AUTO_CLEAR = 0x02



class Max31865():
    rref = 430
    bias = CONFIG_BIAS_ON
    conv = CONFIG_CONV_AUTO
    one_shot = CONFIG_1SHOT_ON
    three_wire = CONFIG_3WIRE_ON
    filter_sel = CONFIG_FILTER_50
        
    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None ):
        self.controller = controller
        self.controller.spi_init( sdi, sdo, sck, cs, width=8, cpol=1, cpha=1 )
        self.config()

    def config( self ):
        cfg = self.bias | self.conv | self.one_shot | self.three_wire | self.filter_sel
        self.writeReg( REG_CONFIG, cfg )

    def writeReg( self, addr, dat ):
        self.controller.spiWrite([(addr&0x7)|0x80, dat]) 

    def readReg( self, addr ):
        return self.controller.spiRead([addr&0x7, 0xFF])[1]

    def readAllRegs( self ):
        ret = {}
        ret['config'] = self.readReg( REG_CONFIG )
        ret['result_h'] = self.readReg( REG_RESULT_H )
        ret['result_l'] = self.readReg( REG_RESULT_L )
        ret['high_fault_h'] = self.readReg( REG_HIGH_FAULT_H )
        ret['high_fault_l'] = self.readReg( REG_HIGH_FAULT_L )
        ret['low_fault_h'] = self.readReg( REG_LOW_FAULT_H )
        ret['low_fault_l'] = self.readReg( REG_LOW_FAULT_L )
        ret['fault'] = self.readReg( REG_FAULT )
        return ret

    def getTemperature( self ):
        ad = (self.readReg( REG_RESULT_H )<<8) | self.readReg( REG_RESULT_L )
        if ad & 0x0001:
            raise Exception("sensor fault")
        r = float(ad>>1) * self.rref / (2**15)
        # calibrate as linear: 203.11 ohm @ 275.0 degree
        # factor = 275.0/(203.11-100) = 2.667 deg/ohm
        t = (r-100)*2.667
        return t




# coding:utf8
__doc__ = 'Vibration Analysis Platform Series'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Mcush


class VAP( Mcush.Mcush ):
    '''VAP 100/200'''
    DEFAULT_NAME = 'VAP'
    DEFAULT_IDN = re_compile( 'VAP[0-9]+.*,([0-9]+\.[0-9]+.*)' )

    def getDeviceName( self ):
        ret = self.writeCommand( '*idn?' )
        try:
            self.device_name = ret[2].strip()
            self.logger.info( 'DeviceName:%s', str(self.device_name) )
            return self.device_name
        except IndexError:
            self.device_name = None
            return ''
        
    def parseEnv( self, info ):
        var = {}
        for l in info:
            ab = l.split(':')
            if len(ab) == 2:
                a, b = ab
            elif len(ab) == 1:
                a, b = ab[0], ''
            var[a.strip()] = b.strip()
        return var
        
    def env( self ):
        return self.parseEnv( self.writeCommand('env') )
 
    def ad( self, command, index=None, value=None ):
        cmd = 'ad7768 -c %s'% command
        if index is not None:
            cmd += ' -i %d'% index
        if value is not None:
            if value & 0xFFFF0000:
                cmd += ' -v 0x%X'% value
            else:
                cmd += ' -v %d'% value
        ret = self.writeCommand( cmd )
        return ret

    def ad_mem( self ):
        a, b = self.ad('mem')[0].split(' ')
        return (int(a,16), int(b,16))
    
    def ad_stop( self ):
        self.ad( 'stop' )


    def daq( self, command, index=None, value=None ):
        cmd = 'daq -c %s'% command
        if index is not None:
            cmd += ' -i %d'% index
        if value is not None:
            if value & 0xFFFF0000:
                cmd += ' -v 0x%X'% value
            else:
                cmd += ' -v %d'% value
        ret = self.writeCommand( cmd )
        return ret

    def daq_lock( self ):
        self.daq('lock')

    def daq_unlock( self ):
        self.daq('unlock')

    def daq_enable( self ):
        self.daq('enable')

    def daq_disable( self ):
        self.daq('disable')

    def daq_busy( self ):
        return bool(int(self.daq('busy')[0]))

    def daq_tacho( self, info=None ):
        val = None
        if info is None:
            info = self.daq('info')
        for l in info:
            if l.startswith('tac:'):
                val = float(l.split(': ')[1].strip())
                break
        return val
    
    def daq_channel_info( self, info=None ):
        ret = {}
        if info is None:
            info = self.daq('info')
        for l in info:
            if l.startswith('#'):
                a, b, c, d = l.split('  ')
                idx = int(a.lstrip('#'))
                cnt = int(b.split(': ')[-1]) 
                avg = float(c.split(': ')[-1]) 
                rms = float(d.split(': ')[-1]) 
                ret[idx] = {'cnt': cnt, 'avg': avg, 'rms': rms}
        return ret
                 



# coding: utf8
__doc__ = 'products from Shenzhen Korad Tech.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
import re
import time
from .. import Env, Instrument


class KA3000( Instrument.SerialInstrument ):
    DEFAULT_CHECK_RETURN_COMMAND = False
    DEFAULT_IDN = re.compile( 'KORAD KA3[0-9]+P V[0-9]+.[0-9]+ SN:[0-9]+' )

    # eg: *IDN?
    #     KORAD KA3005P V5.5 SN:16072670
    
    def __init__( self, *args, **kwargs ):
        kwargs['baudrate'] = 9600  # this is fixed
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 

    def readUntilPrompts( self, line_callback=None ):
        pass

    def connect( self, check_idn=True ):
        Instrument.SerialInstrument.connect( self, check_idn=False )
        idn = self.writeCommand('*IDN?', need_response=True)
        brand, model, ver, sn = idn.split(' ')
        sn = sn.split(':')[1]
        self.model = model
        self.ver = ver
        self.sn = sn
        self.idn = ','.join([model,ver])

    def getSerialNumber( self ):
        return self.sn 

    def writeCommand( self, cmd, need_response=False ):
        # write command and wait for response
        cmd = cmd.strip().upper()
        self.writeLine( cmd )
        self.logger.debug( cmd )
        if need_response:
            line = self.readLine()
            self.logger.debug( line )
            return line

    def output( self, volt=None, amp=None, channel=1, enable=True ):
        if volt is not None:
            volt = float(volt)
            if volt < 0 or volt > self.VMAX:
                raise Exception('Volt out of range')
            self.writeCommand('VSET%d:%.2f'% (channel, volt))
        if amp is not None:
            amp = float(amp)
            if amp < 0 or amp > self.IMAX:
                raise Exception('Current out of range')
            self.writeCommand('ISET%d:%.2f'% (channel, amp))
        if enable is not None:
            self.writeCommand('OUT1' if enable else 'OUT0')
        
    def outputEnable( self ):
        self.output( enable=True )

    def outputDisable( self ):
        self.output( enable=False )

    def getSetting( self, channel=1 ):
        volt = self.writeCommand('VSET%d?'% channel, need_response=True)
        amp = self.writeCommand('ISET%d?'% channel, need_response=True)
        return float(volt), float(amp)
 
    def read( self, channel=1 ):
        volt = self.writeCommand('VOUT%d?'% channel, need_response=True)
        amp = self.writeCommand('IOUT%d?'% channel, need_response=True)
        return float(volt), float(amp)
        
    def beep( self, on=False ):
        self.writeCommand('BEEP%d'% int(on))
        
    def status( self ):
        s = self.writeCommand('STATUS?', need_response=True)
        return ord(s)
        



class KA3005P( KA3000 ):
    DEFAULT_NAME = 'KA3005P'
    VMAX = 30.0
    IMAX = 5.0





# coding: utf8
__doc__ = 'Aosong sensors'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import time


class Dht11():
        
    def __init__( self, controller, pin=None ):
        self.controller = controller
        if pin:
            self.controller.writeCommand('dht11 -p%s'% pin)

    def read( self ):
        ret = self.controller.writeCommand('dht11')
        return float(ret[0]), float(ret[1])


CMD_INIT = 0xE1
CMD_RESET = 0xBA
CMD_CONVERT = 0xAC
CMD_READ = 0x71

class Aht10():
        
    def __init__( self, controller, scl=None, sda=None ):
        self.controller = controller
        self.controller.i2c_init( 0x38, scl, sda )
        self.reset()
        self.calib()

    def calib( self ):
        while not self.readStatus() & 0x08:
            self.controller.i2c( [CMD_INIT, 0x08, 0x00] )
            time.sleep(1)

    def reset( self ):
        self.controller.i2c( [CMD_RESET] )
        time.sleep(0.1)

    def convert( self ):
        self.controller.i2c( [CMD_CONVERT, 0x33, 0x00] )
        time.sleep(0.1)
        while self.readStatus() & 0x80:
            time.sleep(0.1)

    def readStatus( self ):
        ret = self.controller.i2c( [CMD_READ], read=1 )
        return ret[0]

    def read( self ):
        ret = self.controller.i2c( [CMD_READ], read=6 )
        humidity = float((ret[1]<<12)|(ret[2]<<4)|(ret[3]>>4))
        temperature = float(((ret[3]&0xF)<<16)|(ret[4]<<8)|(ret[5]))
        humidity = humidity/(2**20)*100
        temperature = (temperature/(2**20)*200)-50.0
        return humidity, temperature




# coding:utf8
__doc__ = 'qpython on android'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
import os
import re
import sys
import time
import base64
import logging
from .. import Env
from .. import Utils
from .. import Instrument
from .. import Mcush


class QPythonBluetoothPort(Instrument.Port):
        
    UUID = '00001101-0000-1000-8000-00805F9B34FB'

    def __init__( self, parent, *args, **kwargs ):
        Instrument.Port.__init__( self, parent, *args, **kwargs )
        import androidhelper
        self.androidhelper = androidhelper
        self.android = androidhelper.Android()
 
    def connect( self ):
        if self._connected:
            return
        r = self.android.bluetoothConnect(self.UUID, self.port)
        if r.error:
            raise Instrument.SerialNotFound( self.port )
        r = self.android.bluetoothGetConnectedDeviceName('')
        print("%s Connected"% r.result)
        self._connected = True

    def disconnect( self ):
        self.android.bluetoothStop( '' )
        self._connected = False
    
    def read( self, read_bytes=1 ):
        return self.android.bluetoothRead(read_bytes, connID='').result

    def write( self, buf ):
        self.android.bluetoothWrite( buf, '' )
 
    def flush( self ):
        pass



class QMcush( Mcush.Mcush ):
    '''Mcush on Android/QPython'''
    PORT_TYPE = QPythonBluetoothPort

    def tts( self, msg ):
        self.port.android.ttsSpeak( msg )
        while True:
            r = self.port.android.ttsIsSpeaking()
            if r.result:
                time.sleep(0.5)
            else:
                break

    def vibrate( self, duration=300 ):
        self.port.android.vibrate(duration)

    def smsSend( self, addr, msg ):
        self.port.android.smsSend( addr, msg )
        





# coding: utf8
__doc__ = 'Products from Uni-Trend Technology Co., Ltd.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'

from . import UT61




# coding: utf8
__doc__ = 'abstract bitmap canvas with common 2d graphic libs'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from . import Bitmap
from . import Font


# abstract canvas
class Canvas():

    def __init__( self, display ):
        self.display = display
        self.width = display.width
        self.height = display.height

    def flush( self, force=False ):
        # flush screen
        raise NotImplementedError

    def setPixel( self, x, y, color=None, flush=True ):
        # set pixel
        raise NotImplementedError

    def clrScr( self, flush=True ):
        # clear screen
        for y in range(self.height):
            for x in range(self.width):
                self.setPixel( x, y, color=0, flush=False )
        if flush:
            self.flush()

    def drawVLine( self, x, y0, y1, color=None, flush=True ):
        # draw vertical line
        if y1 < y0:
            y0, y1 = y1, y0
        for y in range(y0, y1+1):
            self.setPixel( x, y, color, flush=False )
        if flush:
            self.flush()

    def drawHLine( self, y, x0, x1, color=None, flush=True ):
        # draw horizontal line
        if x1 < x0:
            x0, x1 = x1, x0
        for x in range(x0, x1+1):
            self.setPixel( x, y, color, flush=False )
        if flush:
            self.flush()
 
    def drawRectangle( self, x0, y0, x1, y1, color=None, fill=True, flush=True ):
        # draw rectangle 
        if x0 == x1:
            self.drawVLine( x0, y0, y1, color, flush=False )
        elif y0 == y1:
            self.drawHLine( y0, x0, x1, color, flush=False )
        else:
            self.drawHLine( y0, x0, x1, color, flush=False )
            self.drawHLine( y1, x0, x1, color, flush=False )
            if fill:
                for y in range(y0+1, y1):
                    self.drawHLine( y, x0, x1, color, flush=False )
            else:
                self.drawVLine( x0, y0, y1, color, flush=False )
                self.drawVLine( x1, y0, y1, color, flush=False )
        if flush:
            self.flush()
    
    def drawLine( self, x0, y0, x1, y1, color=None, flush=True ):
        # draw normal line
        if x1 < x0:
            x0, y0, x1, y1 = x1, y1, x0, y0
        if x0 == x1:
            self.drawVLine( x0, y0, y1, color, flush=False )
        elif y0 == y1:
            self.drawHLine( y0, x0, x1, color, flush=False )
        else:
            self.setPixel( x0, y0, color, flush=False )
            self.setPixel( x1, y1, color, flush=False )
            k = float(y1-y0)/(x1-x0)
            if abs(k) < 1.0:
                for x in range(x0+1, x1):
                    y = int(y0+(x-x0)*k+0.5)
                    self.setPixel( x, y, color, flush=False )
            else:
                if k > 0:
                    for y in range(y0+1, y1):
                        x = int(x0+(y-y0)/k+0.5)
                        self.setPixel( x, y, color, flush=False )
                else:
                    for y in range(y0-1, y1, -1):
                        x = int(x0+(y-y0)/k+0.5)
                        self.setPixel( x, y, color, flush=False )
        if flush:
            self.flush()
 
    def drawBitmap( self, x, y, bitmap, mode='normal', flush=True ):
        # draw bitmap
        for i in range(bitmap.height):
            for j in range(bitmap.width):
                c = bitmap.getPixel(j, i)
                if mode == 'normal':
                    self.setPixel( x+j, y+i, int(c), flush=False )
                elif mode == 'transparent':
                    if c:
                        self.setPixel( x+j, x+i, flush=False )
        if flush: 
            self.flush() 

    def drawString( self, x, y, string, mode='normal', font=None, flush=True ):
        # draw text
        if font is None:
            font = Font.DEFAULT_FONT
        for s in string:
            if not s in font:
                s = None
            elif font[s] is None:
                s = '?'
            bitmap = font[s]
            self.drawBitmap( x, y, bitmap, mode, flush=False )
            x += bitmap.width
            if x >= self.width:
                break
        if flush: 
            self.flush() 

    def getStringRenderSize( self, string, font=None ):
        # calculate string render size
        width = 0
        height = 0
        if font is None:
            font = Font.DEFAULT_FONT
        for s in string:
            if not s in font:
                s = None
            elif font[s] is None:
                s = '?'
            bitmap = font[s]
            w, h = bitmap.width, bitmap.height
            width += w
            if h > height:
                height = h
        return (width, height) 




# coding: utf8
__doc__ = 'basic environment'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from os import getenv
from os.path import isdir, join
from sys import platform, version_info
from binascii import unhexlify
from tempfile import mktemp
from subprocess import check_output


_bool_true_list = ['1', 'Y', 'y', 'T', 't', 'yes', 'Yes', 'YES', 'true', 'True', 'TRUE']

 
def getenv_bool( key, default=None ):
    ret = getenv(key, default)
    return False if ret is None else bool(ret in _bool_true_list)
 
def getenv_float( key, default=None ):
    ret = getenv(key, default)
    return None if ret is None else float(ret)
    
def getenv_int( key, default=None ):
    ret = getenv(key, default)
    try:
        if isinstance(ret, str):
            return int(eval(ret))
        else:
            return int(ret)
    except:
        return None
    #return None if ret is None else int(ret)

TEST_DIR = getenv('TEST_DIR')
if not TEST_DIR:
    if platform == 'linux2' and isdir( '/dev/shm' ):
        TEST_DIR = '/dev/shm'
    else:
        from tempfile import gettempdir
        TEST_DIR = gettempdir()

TEST_PREFIX = getenv('TEST_PREFIX', 'mcush_test' )
TEST_WR = mktemp( prefix=TEST_PREFIX + "_wr_", dir=TEST_DIR )
TEST_RD = mktemp( prefix=TEST_PREFIX + "_rd_", dir=TEST_DIR )
TEST_MINID = getenv_int('TEST_MINID', 0)
TEST_MAXID = getenv_int('TEST_MAXID', 100)
TEST_MAXSIZE = getenv_int('TEST_MAXSIZE', 10*1024)


if platform == 'win32':
    PORT = getenv('PORT', 'COM1')
else:
    PORT = getenv('PORT', '/dev/ttyUSB0')
    try:
        PORTS = check_output(['allports']).strip().decode(encoding='utf8')
        PORTS_LIST = PORTS.split(',')
    except:
        PORTS_LIST = []
BAUDRATE = getenv_int( 'BAUDRATE', 9600 )
RTSCTS = getenv_bool( 'RTSCTS' )
PARITY = getenv( 'PARITY', 'N' )
TIMEOUT = getenv_int('TIMEOUT', 5)
COMMAND_FAIL_RETRY = getenv_int( 'COMMAND_FAIL_RETRY', 3 )

DELAY = getenv_float( 'DELAY', 1 )
DELAY_AFTER_REBOOT = getenv_float( 'DELAY_AFTER_REBOOT', 1 )
RETRY = getenv_int( 'RETRY', 1000 )
NO_ECHO_CHECK = getenv_bool( 'NO_ECHO_CHECK' )
NO_IDN_CHECK = getenv_bool( 'NO_IDN_CHECK' )
COMPACT_MODE = getenv_bool( 'COMPACT_MODE' )

DEVELOPMENT = getenv_bool( 'DEVELOPMENT' )
PROFILE = getenv_bool( 'PROFILE' )
LOGGING_FORMAT = getenv('LOGGING_FORMAT')

DEBUG = getenv_bool( 'DEBUG' )
VERBOSE = getenv_bool( 'VERBOSE' )
INFO = getenv_bool( 'INFO' )
WARNING = getenv_bool( 'WARNING' )


ESPEAK = getenv_bool( 'ESPEAK' )

PDF_READER = None
SOFFICE_BIN = None 
if platform == 'win32':
    PDF_READER = 'AcroRd32.exe'
    SOFFICE_BIN = 'soffice.exe'
    MPLAYER_BIN = 'mplayer.exe'
    NOTEPAD_BIN = 'notepad.exe'
elif platform == 'linux2':
    PDF_READER = '/usr/bin/evince'
    SOFFICE_BIN = '/usr/bin/soffice'
    MPLAYER_BIN = '/usr/bin/mplayer'
    NOTEPAD_BIN = '/usr/bin/mousepad'

REV = getenv_bool( 'REV' )
SIM = getenv_bool( 'SIM' )

LANGUAGES = {'en', 'zh_cn'}
LANGUAGE = getenv( 'LANGUAGE', 'en' ).split(':')[0].lower()

PYTHON_V3 = bool(version_info > (3, 0))
if PYTHON_V3:
    EMPTY_BYTE = b''
else:
    EMPTY_BYTE = ''

LOG_DATEFMT = '%Y-%m-%d %H:%M:%S'
LOG_FORMAT = '%(asctime)s %(name)s %(levelname)s %(message)s'
 
try:
    if PYTHON_V3:
        from mcush.EnvExtra import *
    else:
        from EnvExtra import *
except ImportError:
    pass




# coding:utf8
__doc__ = 'Shell Lab Series'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Mcush, Instrument, Utils
import binascii
import time


class ShellLab(Mcush.Mcush):
    DEFAULT_NAME = 'ShellLab'
    DEFAULT_IDN = re_compile( 'ShellLab(-[A-Z][0-9a-zA-Z\-]*)?,([0-9]+\.[0-9]+.*)' )

    def scpiRst( self ):
        if self.checkCommand("*rst"):
            self.writeCommand('*rst')
        else:
            self.errno( 0 )
            # clear all leds
            for i in range(self.getLedNumber()):
                self.led(i, False)

    def daq( self, cmd, index=None, value=None ):
        cmd = 'daq -c %s'% cmd
        if index is not None:
            cmd += ' -i %d'% index
        if value is not None:
            cmd += ' -v %d'% value
        return self.writeCommand( cmd )

    def daq_init( self, freq=None, length=None, channels=[], channel_mask=None ):
        if freq is not None:
            self.daq( 'freq', value=freq )
        if length is not None:
            self.daq( 'length', value=length )
        if channel_mask is not None:
            self.daq( 'channel_mask', value=channel_mask )
            channels = []
            for i in range(8):
                if channel_mask & (1<<i):
                    channels.append(i)
            self.channels = channels
        else:
            mask = 0
            for c in channels:
                mask |= 1<<c
            self.daq( 'channel_mask', value=mask )
            self.channels = channels
        self.daq( 'init' )
        self.vref = float(self.daq('vref')[0])

    def daq_deinit( self ):
        self.daq( 'deinit' )

    def daq_start( self ):
        self.daq( 'start' )

    def daq_stop( self ):
        self.daq( 'stop' )

    def daq_reset( self ):
        self.daq( 'reset' )

    def daq_done( self ):
        return int(self.daq( 'done' )[0])

    def daq_read( self, channel ):
        ret = self.daq( 'read', index=channel )
        dat = []
        for l in ret:
            for v in l.strip().split(','):
                dat.append( int(v, 16) * self.vref / 4096 )
        return dat

    daqInit = daq_init
    daqDeinit = daq_deinit
    daqStart = daq_start
    daqStop = daq_stop
    daqReset = daq_reset
    daqDone = daq_done
    daqRead = daq_read

    def measure( self, cmd, index=None, value=None ):
        cmd = 'measure -c %s'% cmd
        if index is not None:
            cmd += ' -i %d'% index
        if value is not None:
            cmd += ' -v %d'% value
        return self.writeCommand( cmd )

    def measure_start( self ):
        self.measure( 'start' )

    def measure_stop( self ):
        self.measure( 'stop' )

    def measure_read( self, channel ):
        ret = self.measure( 'read', index=channel )
        dat = []
        for l in ret:
            for v in l.strip().split(','):
                dat.append( float(v) )
        return dat

    measureStart = measure_start
    measureStop = measure_stop
    measureRead = measure_read


COLOR_TAB = {
'black'  : 0x000000,
'red'    : 0xFF0000,
'green'  : 0x00FF00,
'blue'   : 0x0000FF,
'yellow' : 0xFFFF00,
'cyan'   : 0x00FFFF,
'purple' : 0xFF00FF,
'white'  : 0xFFFFFF,
}


class ShellLabLamp(Mcush.Mcush):
    DEFAULT_NAME = 'ShellLabLamp'
    DEFAULT_IDN = re_compile( 'ShellLab-L1[a-zA-Z]*,([0-9]+\.[0-9]+.*)' )

    def lamp( self, color=None, red=None, green=None, blue=None, freq=None ):
        cmd = 'lamp'
        if color is not None:
            cmd += ' -c 0x%X'% color
        if red is not None:
            cmd += ' -r %d'% red
        if green is not None:
            cmd += ' -g %d'% green
        if blue is not None:
            cmd += ' -b %d'% blue
        if freq is not None:
            cmd += ' -f %s'% freq
        return self.writeCommand(cmd)

    def alarm( self, count=None, freq=None ):
        cmd = 'alarm'
        if count is not None:
            cmd += ' -c %d'% count
        if freq is not None:
            cmd += ' -f %d'% freq
        return self.writeCommand(cmd)
 
    def reset( self, lamp_freq=1, alarm_freq=1 ):
        self.lamp( 0, freq=lamp_freq )
        self.alarm( 0, freq=alarm_freq )

    def color( self, c, freq=None ):
        if isinstance(c, str):
            if not c in COLOR_TAB:
                raise Exception('Unknown color name %s'% c)
            c = COLOR_TAB[c]
        self.lamp( c, freq=freq ) 


class ShellLabStrap(Mcush.Mcush):
    DEFAULT_NAME = 'ShellLabStrap'
    DEFAULT_IDN = re_compile( 'ShellLab-L2[a-zA-Z]*,([0-9]+\.[0-9]+.*)' )

    def __init__( self, *args, **kwargs ):
        Instrument.Instrument.__init__( self, *args, **kwargs )
        self.length = kwargs.get('length', None)
        if self.length is not None:
            self.strapLength( self.length ) 

    def strapLength( self, length ):
        cmd = 'strap -l%d'% length
        self.writeCommand(cmd)

    def strap( self, color=None, red=None, green=None, blue=None, freq=None ):
        cmd = 'strap'
        if color is not None:
            cmd += ' -c 0x%X'% color
        if red is not None:
            cmd += ' -r %d'% red
        if green is not None:
            cmd += ' -g %d'% green
        if blue is not None:
            cmd += ' -b %d'% blue
        if freq is not None:
            cmd += ' -f %s'% freq
        return self.writeCommand(cmd)

    def reset( self, freq=1 ):
        self.strap( 0, freq=freq )

    def color( self, c, freq=None ):
        if isinstance(c, str):
            if not c in COLOR_TAB:
                raise Exception('Unknown color name %s'% c)
            c = COLOR_TAB[c]
        self.strap( c, freq=freq ) 

    def beep( self, freq=None, duration=0.05, times=1 ):
        # not supported
        pass

    
class ShellLabStaticStrap(ShellLabStrap):
    def __init__( self, *args, **kwargs ):
        ShellLabStrap.__init__(self, *args, **kwargs)
        setting = Utils.parseKeyValueLines(self.strap())
        if float(setting['freq']) > 0:
            self.strap( freq=0 )
        self._color = int(setting['color'], 16)
        self._brightness = kwargs.get('brightness', 1.0)
        
    def reset( self ):
        self.setColor( 0 )

    def calcDimColor( self, c ):
        r, g, b = (c>>16)&0xFF, (c>>8)&0xFF, c&0xFF
        r = int(r*self._brightness)
        g = int(g*self._brightness)
        b = int(b*self._brightness)
        return ((r&0xFF)<<16)|((g&0xFF)<<8)|(b&0xFF)

    def color( self, c, brightness=None ):
        self._color = c
        if brightness is not None:
            self._brightness = brightness
        if c in COLOR_TAB:
            c = COLOR_TAB[c]
        self.strap(self.calcDimColor(c))

    def brightness( self, b ):
        self._brightness = b
        c = self._color
        if c in COLOR_TAB:
            c = COLOR_TAB[c]
        self.strap(self.calcDimColor(c))
            
    setColor = color
    setBrightness = brightness
   

class ShellLabCAN(ShellLab):
    DEFAULT_NAME = 'ShellLabCAN'
    DEFAULT_IDN = re_compile( 'ShellLab-C[0-9a-zA-Z\-]*,([0-9]+\.[0-9]+.*)' )

# CANopen function codes
NMT   = 0x0<<7
SYNC  = 0x1<<7
TIME  = 0x2<<7
TPDO1 = 0x3<<7
RPDO1 = 0x4<<7
TPDO2 = 0x5<<7
RPDO2 = 0x6<<7
TPDO3 = 0x7<<7
RPDO3 = 0x8<<7
TPDO4 = 0x9<<7
RPDO4 = 0xA<<7
TSDO  = 0xB<<7
RSDO  = 0xC<<7
NDGRD = 0xE<<7
LSS   = 0xF<<7
# NMT commands
NMT_START = 1
NMT_STOP = 2
NMT_PRE_OPER = 0x80
NMT_RESET = 0x81
NMT_RESET_COMM = 0x82
# NODE status
STATUS_BOOT_UP = 0
STATUS_CONNECTING = 2
STATUS_STOPPED = 4
STATUS_OPERATIONAL = 5
STATUS_PRE_OPERATIONAL = 0x7F

def _v2l( val ):
    if isinstance(val, list):
        return val
    else:
        return map(ord,val)[:8]

class ShellLabCANopen(ShellLabCAN):
    DEFAULT_SDO_TIMEOUT = 5

    # SYNC command
    def sync( self ):
        self.canWrite( SYNC )

    # NODE_GUARD command
    def writeNodeGuardRequest( self, id ):
        self.canWrite( NDGRD+(id&0x7F), rtr=True )

    # NMT command
    def writeNMTCommand( self, cmd, id ):
        self.canWrite( NMT, [cmd, id&0x7F] )
        
    def resetNode( self, id ):
        self.writeNMTCommand( NMT_RESET, id )

    def startNode( self, id ):
        self.writeNMTCommand( NMT_START, id )
 
    def stopNode( self, id ):
        self.writeNMTCommand( NMT_STOP, id )

    def preOperNode( self, id ):
        self.writeNMTCommand( NMT_PRE_OPER, id )

    # RPDO commands 
    def writeRPDO1( self, id, val ):
        self.canWrite( RPDO1+(id&0x7F), _v2l(val) )

    def writeRPDO2( self, id, val ):
        self.canWrite( RPDO2+(id&0x7F), _v2l(val) )

    def writeRPDO3( self, id, val ):
        self.canWrite( RPDO3+(id&0x7F), _v2l(val) )

    def writeRPDO4( self, id, val ):
        self.canWrite( RPDO4+(id&0x7F), _v2l(val) )

    # TPDO commands 
    def writeTPDO1Request( self, id ):
        self.canWrite( TPDO1+(id&0x7F), rtr=True  )

    def writeTPDO2Request( self, id ):
        self.canWrite( TPDO2+(id&0x7F), rtr=True  )

    def writeTPDO3Request( self, id ):
        self.canWrite( TPDO3+(id&0x7F), rtr=True  )

    def writeTPDO4Request( self, id ):
        self.canWrite( TPDO4+(id&0x7F), rtr=True  )

    # Object dictionaries
    def readObject( self, id, index, subindex ):
        self.canWrite( RSDO+id, [0x5F, index&0xFF, (index>>8)&0xFF, subindex, 0, 0, 0, 0] )
        bytes_required = None
        t0 = time.time()
        responsed = None
        while not responsed:
            if time.time() > t0 + self.DEFAULT_SDO_TIMEOUT:
                raise Exception("SDO upload request timeout")
            for cid, ext, rtr, dat in self.canRead(block_ms=0):
                #self.logger.debug( 'id=0x%X, ext=%d, rtr=%d, dat=%s'% (cid, ext, rtr, binascii.hexlify(dat)) )
                if (cid != TSDO+id) or rtr or (len(dat) != 8):
                    continue
                if (Utils.s2H(dat[1:3]) != index) or (Utils.s2B(dat[3]) != subindex):
                    continue
                m = Utils.s2B(dat[0])
                transfer_type = bool(m & 0x02)
                size_indicator = bool(m & 0x01)
                #self.logger.debug( 'transfer_type: %d  size_indicator: %d'% (transfer_type, size_indicator) )
                if transfer_type:
                    # expedited
                    not_used = (m>>2)&3
                    #self.logger.debug( 'SDO read %d bytes'% (4-not_used) )
                    if size_indicator and not_used:
                        return dat[4:-not_used]
                    else:
                        return dat[4:]
                else:
                    # normal
                    if size_indicator:
                        bytes_required = Utils.s2I(dat[4:])
                    else:
                        bytes_required = 0xFFFFFF  # unknown length
                    #self.logger.debug( 'SDO bytes_required=%X'% (bytes_required) )
                    responsed = True
                    break
        if not bytes_required:
            return
        # continue to read
        read = []
        read_bytes = 0
        control = 0x60
        while read_bytes < bytes_required:
            #self.logger.debug( 'SDO Segment Req control=%X'% (control) )
            self.canWrite( RSDO+id, [control, 0, 0, 0, 0, 0, 0, 0] )
            t0 = time.time()
            responsed = None
            while not responsed:
                if time.time() > t0 + self.DEFAULT_SDO_TIMEOUT:
                    raise Exception("SDO upload segment Timeout")
                for cid, ext, rtr, dat in self.canRead(block_ms=0):
                    #self.logger.debug( 'SDO Segment id=0x%X, ext=%d, rtr=%d, dat=%s'% (cid, ext, rtr, binascii.hexlify(dat)) )
                    if (cid==(TSDO+id)) and (not rtr) and (len(dat)==8):
                        m = Utils.s2B(dat[0])
                        if m & 0x80:
                            raise Exception("SDO Abort")
                        not_used = (m>>1)&7
                        no_more = m&1
                        if not_used:
                            read.append( dat[1:-not_used] )
                        else:
                            read.append( dat[1:] )
                        read_bytes += 7-not_used
                        if no_more:
                            bytes_required = read_bytes
                        #self.logger.debug( 'SDO read_len=%s, no_more=%d'% (read_bytes, no_more) )
                        responsed = True
                        break
            control ^= 0x10
        return ''.join(read)
     
    def writeObject( self, id, index, subindex, val ):
        val_len = len(val)
        control = 0x20
        bytes_written = 0
        if val_len <= 4:
            # expedited mode
            control |= 0x02
            v = map(ord, val)
            if len(v) < 4:
                control |= 0x01  # indicated 
                control |= (4-len(v)) << 2 # not used bytes
                while len(v) < 4:
                    v.append( 0 )
            self.canWrite( RSDO+id, [control, index&0xFF, (index>>8)&0xFF, subindex] + v )
            bytes_written = val_len
        else:
            # normal mode
            control |= 0x01  # indicated 
            self.canWrite( RSDO+id, [control, index&0xFF, (index>>8)&0xFF, subindex, 
                           val_len&0xFF, (val_len>>8)&0xFF, (val_len>>16)&0xFF, 0] )
        # initial download response
        t0 = time.time()
        responsed = None
        while not responsed:
            if time.time() > t0 + self.DEFAULT_SDO_TIMEOUT:
                raise Exception("SDO download request timeout")
            for cid, ext, rtr, dat in self.canRead(block_ms=0):
                #self.logger.debug( 'id=0x%X, ext=%d, rtr=%d, dat=%s'% (cid, ext, rtr, binascii.hexlify(dat)) )
                if (cid!=TSDO+id) or rtr or (len(dat)!=8):
                    continue
                m = Utils.s2B(dat[0])
                if m & 0x80:
                    raise Exception("SDO Abort")
                if (Utils.s2H(dat[1:3]) != index) or (Utils.s2B(dat[3]) != subindex):
                    continue
                responsed = True
                break
        # remaining segment request
        control = 0x00
        while bytes_written < val_len:
            v = map(ord, val[bytes_written:bytes_written+7])
            bytes_written += len(v)
            if bytes_written >= val_len:
                control |= 0x01  # no more flag
            if len(v) < 7:
                control |= (7-len(v))<<1  # not used bytes
                while len(v) < 7:
                    v.append( 0 )
            self.canWrite( RSDO+id, [control] + v )
            t0 = time.time()
            responsed = None
            while not responsed:
                if time.time() > t0 + self.DEFAULT_SDO_TIMEOUT:
                    raise Exception("SDO download segment timeout")
                for cid, ext, rtr, dat in self.canRead(block_ms=0):
                    #self.logger.debug( 'id=0x%X, ext=%d, rtr=%d, dat=%s'% (cid, ext, rtr, binascii.hexlify(dat)) )
                    if (cid!=TSDO+id) or rtr or (len(dat)!=8):
                        continue
                    m = Utils.s2B(dat[0])
                    if m & 0x80:
                        raise Exception("SDO Abort")
                    responsed = True
                    break
            control ^= 0x10
        
    def readUINT32( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2I(val[:4])

    def readINT32( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2i(val[:4])

    def readUINT16( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2H(val[:2])

    def readINT16( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2h(val[:2])

    def readUINT8( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2B(val[:1])

    def readINT8( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2b(val[:1])

    def readFLOAT32( self, id, index, subindex ):
        val = self.readObject( id, index, subindex )
        return Utils.s2f(val[:4])

    def writeUINT32( self, id, index, subindex, val ):
        dat = Utils.I2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeINT32( self, id, index, subindex, val ):
        dat = Utils.i2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeUINT16( self, id, index, subindex, val ):
        dat = Utils.H2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeINT16( self, id, index, subindex, val ):
        dat = Utils.h2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeUINT8( self, id, index, subindex, val ):
        dat = Utils.B2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeINT8( self, id, index, subindex, val ):
        dat = Utils.b2s(val)
        self.writeObject( id, index, subindex, dat )

    def writeFLOAT32( self, id, index, subindex, val ):
        dat = Utils.f2s(val)
        self.writeObject( id, index, subindex, dat )







# coding: utf8
__doc__ = 'MCUSH module'
__version__ = '1.4.8'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__website__ = 'https://github.com/pengshulin/mcush/'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'

from . import Env
from . import Utils
from . import Instrument
from . import Register
from . import Mcush
from . import AppUtils

from .android import *
from .misc import *
from .linkong import *
from .fluke import *
from .uni_trend import *
from .rigol import *



# coding: utf8
__doc__ = 'Products from Beijing RIGOL Technologies Inc.'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'

from . import DP700
from . import DP800
from . import DG800
from . import DM3058
from . import DS1000




# coding:utf8
__doc__ = 'UT61B/C/D/E digital multimeter'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env

FUNCTIONS = {
0x3B: 'Voltage',
0x3D: 'Auto uA Current',
0x3F: 'Auto mA Current',
0x30: '22A Current',
0x39: 'Manual A Current',
0x33: 'Ohm',
0x35: 'Continuity',
0x31: 'Diode',
0x32: 'Frequency',
0x36: 'Capacitance',
0x34: 'Temperature',
0x3E: 'ADP',
}


RANGE_VOLT = {
0x30: (2.2, 0.0001, 'V'),
0x31: (22.0, 0.001, 'V'),
0x32: (220.0, 0.01, 'V'),
0x33: (2200.0, 0.1, 'V'),
0x34: (220.0, 0.01, 'mV'),
}

RANGE_AUTO_UAMP = {
0x30: (2.20, 0.01, 'uA'),
0x31: (22.0, 0.1, 'uA'),
}

RANGE_AUTO_MAMP = {
0x30: (2.200, 0.001, 'mA'),
0x31: (22.00, 0.01, 'mA'),
}

RANGE_22AMP = {
0x30: (22.0, 0.001, 'A'),
}

RANGE_220AMP = {
0x30: (22.0, 0.001, 'A'),
0x31: (220.0, 0.01, 'A'),
}

RANGE_2200AMP = {
0x30: (220.0, 0.01, 'A'),
0x31: (2200.0, 0.1, 'A'),
}


RANGE_MANUAL_AMP = {
0x30: (2.2, 0.0001, 'A'),
0x31: (22.0, 0.001, 'A'),
0x32: (220.0, 0.01, 'A'),
0x33: (2200.0, 0.1, 'A'),
0x34: (22000.0, 1.0, 'A'),
}

RANGE_OHM = {
0x30: (220.0, 0.01, 'Ohm'),
0x31: (2.2, 0.0001, 'kOhm'),
0x32: (22.0, 0.001, 'kOhm'),
0x33: (220.0, 0.01, 'kOhm'),
0x34: (2.2, 0.0001, 'MOhm'),
0x35: (22.0, 0.001, 'MOhm'),
0x36: (220.0, 0.01, 'MOhm'),
}

RANGE_FREQ = {
0x30: (22.0, 0.01, 'Hz'),
0x31: (220.0, 0.1, 'Hz'),
0x33: (22.0, 0.001, 'kHz'),
0x34: (220.0, 0.01, 'kHz'),
0x35: (2.2, 0.0001, 'MHz'),
0x36: (22.0, 0.001, 'MHz'),
0x37: (220.0, 0.01, 'MHz'),
}

RANGE_CAP = {
0x30: (22.0, 0.001, 'nF'),
0x31: (220.0, 0.01, 'nF'),
0x32: (2.2, 0.0001, 'uF'),
0x33: (22.0, 0.001, 'uF'),
0x34: (220.0, 0.01, 'uF'),
0x35: (2.2, 0.0001, 'mF'),
0x36: (22.0, 0.001, 'mF'),
0x37: (220.0, 0.01, 'mF'),
}

FORMATTER = {
1.0: "%.0f",
0.1: "%.1f",
0.01: "%.2f",
0.001: "%.3f",
0.0001: "%.4f",
0.00001: "%.5f",
}


class UT61E( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'UT61'
    DEFAULT_PROMPTS = re_compile( '\x0A' )
    DEFAULT_IDN = re_compile( '.*' )

    def __init__( self, *args, **kwargs ):
        kwargs['baudrate'] = 19200  # this is fixed
        kwargs['check_idn'] = False  # not support scpi commands
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 

    def connect( self, check_idn=False ):
        self.port.connect()
        if not self.port.connected:
            return
        self.port.ser.setRTS( 0 )  # power up the opto-isolated circuit
        #self.readUntilPrompts()

    def writeCommand( self, cmd ):
        # UT61X only receives serial packets
        raise Exception("Not supported")

    def formatReading( self, reading, precision ):
        return FORMATTER[precision]% reading

    def getReading( self ):
        # read measure value, function, range, flags
        while True:
            line = self.readLine( eol='\x0D', timeout=self.DEFAULT_TIMEOUT )
            if len(line) == 0:
                #raise Exception("Timeout")
                raise Instrument.CommandTimeoutError()
            if (len(line) != 13) or ((ord(line[0])&0x7F) != 0x0A):
                continue  # ignore invalid packet
            # parse packet
            function = ord(line[7]) & 0x7F 
            if function in FUNCTIONS:
                function = FUNCTIONS[function]
            else:
                continue  # ignore invalid function
            rng = ord(line[1]) & 0x7F 
            reading = ''.join([chr(ord(d)&0x7F) for d in line[2:7]])
            status = ord(line[8]) & 0x7F 
            opt1 = ord(line[9]) & 0x7F 
            opt2 = ord(line[10]) & 0x7F 
            opt3 = ord(line[11]) & 0x7F 
            opt4 = ord(line[12]) & 0x7F 
            # search for valid range list 
            rng_lst = None
            vahz = opt3 & 0x01
            if vahz:
                rng_list = RANGE_FREQ
            elif function == 'Voltage':
                rng_list = RANGE_VOLT
            elif function == 'Auto uA Current':
                if opt4 & 0x04:
                    rng_list = RANGE_220AMP
                else:
                    rng_list = RANGE_AUTO_UAMP
            elif function == 'Auto mA Current':
                if opt4 & 0x04:
                    rng_list = RANGE_2200AMP
                else:
                    rng_list = RANGE_AUTO_MAMP
            elif function == '22A Current':
                rng_list = RANGE_22AMP
            elif function == 'Manual A Current':
                rng_list = RANGE_MANUAL_AMP
            elif function in ['Ohm', 'Continuity', 'Diode']:
                rng_list = RANGE_OHM
            elif function == 'Frequency':
                rng_list = RANGE_FREQ
            elif function == 'Capacitance':
                rng_list = RANGE_CAP
            try:
                full_range, precision, unit = rng_list[rng]
            except:
                continue  # ignore invalid range
            try:
                reading = int(reading) * precision
            except:
                continue  # ignore invalid reading 
            # many flag bits (not all)
            flags = []
            if status & 0x4: # sign bit
                reading = -reading
                flags.append("SIGN")
            if status & 0x1:
                flags.append("OL")
            if opt1 & 0x02:
                flags.append("REL")
            if opt3 & 0x08:
                flags.append("DC")
            if opt3 & 0x04:
                flags.append("AC")
            if opt3 & 0x02:
                flags.append("AUTO")
            if opt3 & 0x01:
                flags.append("VAHZ")
            if opt4 & 0x02:
                flags.append("HOLD")
            return function, reading, precision, unit, flags





# coding: utf8
__doc__ = 'HX711 24-bit ADC'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from .. import Mcush, Utils
import time


MODE_A128 = 0
MODE_B32 = 1
MODE_A64 = 2


class Hx711():
    ref = 5.0
    mode = MODE_A128
    clocks = 25

    def __init__( self, controller, sdi=None, sdo=None, sck=None, cs=None, mode='a128', ref=5.0 ):
        self.controller = controller
        self.sdi = sdi if sdi is not None else '0.0'
        self.sdo = sdo if sdo is not None else '0.1'
        self.sck = sck if sck is not None else '0.2'
        self.cs  = cs  if cs  is not None else '0.3'
        self.ref = ref
        self._setMode( mode )
        self.controller.spiInit(self.sdi, self.sdo, self.sck, self.cs, width=self.clocks, cpha=True)

    def _setMode( self, mode ):
        if mode in [0,'a128','A128','a','A']:
            self.mode = MODE_A128
        elif mode in [1,'b32','B32','b','B']:
            self.mode = MODE_B128
        elif mode in [0,'a64','A64']:
            self.mode = MODE_A64
        else:
            raise Exception('invalid mode')
        self.clocks = self.mode + 25

    def setMode( self, mode ):
        self._setMode( mode )
        self.controller.spiUpdate(self.sdi, self.sdo, self.sck, self.cs, width=self.clocks, cpha=True)
 
    def dataIsReady( self ):
        return self.controller.pinIsLow(self.sdi)

    def readData( self, mode=None ):
        a = self.controller.spiRead([0])[0]
        if mode == 'hex':
            return a>>(self.mode+1)
        else:
            val = Utils.I2i(a<<(7-self.mode))/256
            if mode == 'int':
                return val
            v = self.ref * val / (2**24)
            # return voltage from gain setting
            if self.mode == MODE_A128:
                return v / 128
            elif self.mode == MODE_B32:
                return v / 32
            elif self.mode == MODE_A64:
                return v / 64




# coding:utf8
__doc__ = 'DS1000B/E/U/Z Digital Oscilloscope'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env
import time

# Product lines:
#            Bandwidth   Channels  Samplerate    Depth
# DS1000E/U  50M ~ 100M     2       1G Sa/s      1M pts
# DS1000Z    50M ~ 100M     4       1G Sa/s     24M pts
# DS1000Z-E  200M           2       1G Sa/s     24M pts
# DS1000B    70M ~ 200M     4       2G Sa/s     16k pts 

B2S = {0:'OFF', 1:'ON', True:'ON', False:'OFF', None:'OFF'}

class DS1000( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'DS1000'
    DEFAULT_TERMINATOR_RESET = ''
    DEFAULT_IDN = re_compile( 'Rigol Technologies,DS1.*' )
    DEFAULT_CHECK_RETURN_COMMAND = False

    # eg: *IDN?
    #     Rigol Technologies,DS1102E,DS1ET162359072,00.04.01.00.02

    def readUntilPrompts( self, line_callback=None ):
        pass

    def writeCommand( self, cmd, need_response=False ):
        # write command and wait for response
        cmd = cmd.strip().upper()
        self.writeLine( cmd )
        self.logger.debug( cmd )
        time.sleep(0.2)
        if need_response:
            line = self.readLine()
            self.logger.debug( line )
            return line

    def scpiIdn( self ):
        self.idn = self.writeCommand('*idn?', need_response=True)
        self.brand, self.model, self.sn, self.ver = self.idn.split(',')
        return self.idn

    def scpiRst( self ):
        Instrument.SerialInstrument.scpiRst(self)
        time.sleep(0.5)
 
    def getSerialNumber( self ):
        return self.scpiIdn().split(',')[-2]

    def run( self ):
        # this command does not work, use key('run') instead
        self.writeCommand(":SYST:RUN")
 
    def stop( self ):
        self.writeCommand(":SYST:STOP")
 
    def auto( self ):
        self.writeCommand(":SYST:AUTO")
 
    def snapshot( self ):
        self.writeCommand(":SYST:HARD")

    def beepEnable( self, enable=True ):
        self.writeCommand(":BEEP:ENABLE %s"% B2S[enable])
        
    def beep( self ):
        self.writeCommand(":BEEP:ACT")

    def key( self, key_name ):
        self.writeCommand(":KEY:%s"% key_name) 

    def display( self, disp_type='VECT', grid_type='HALF', persist=False ):
        self.writeCommand(":DISP:TYPE %s"% disp_type)
        self.writeCommand(":DISP:GRID %s"% grid_type)
        self.writeCommand(":DISP:PERS %s"% B2S[persist])
        self.writeCommand(":DISP:CLE")

    def timebase( self, format='YT', delay=False, offset=0, scale=1 ):
        self.writeCommand(":TIM:FORMAT %s"% format )
        self.writeCommand(":TIM:MODE %s"% ('del' if delay else'main' ))
        self.writeCommand(":TIM:OFFS %s"% offset )
        self.writeCommand(":TIM:SCAL %s"% scale )

    def trigger( self, mode, source=None, level=None, sweep=None ):
        # mode type: edge/pulse/video/slope/pattern/duration/alternation
        mode = mode.upper()[:4]
        self.writeCommand(":TRIG:MODE %s"% mode )
        if source is not None:
            self.writeCommand(":TRIG:%s:SOUR %s"% (mode, source) )
        if level is not None:
            self.writeCommand(":TRIG:%s:LEV %s"% (mode, level) )
        if sweep is not None:
            self.writeCommand(":TRIG:%s:SWE %s"% (mode, sweep) )
        
    def forceTrig( self ):
        self.writeCommand(":FORC" )

    def channel( self, channel_idx=1, display=True, coupling='DC',
                       scale=1, offset=0, bw_limit=False, invert=False, 
                       probe=1, filter_on=False ):
        prefix = ':CHAN%d:'% channel_idx
        self.writeCommand(prefix+'DISP %s'% B2S[display])
        self.writeCommand(prefix+'COUP %s'% coupling)
        self.writeCommand(prefix+'SCAL %s'% scale)
        self.writeCommand(prefix+'OFFS %s'% offset)
        self.writeCommand(prefix+'PROB %s'% probe)
        self.writeCommand(prefix+'INV %s'% B2S[invert])
        self.writeCommand(prefix+'BWL %s'% B2S[bw_limit])
        self.writeCommand(prefix+'FILT %s'% B2S[filter_on])

    def waveform( self, mode=None, channel_idx=1 ):
        if mode is not None:
            # mode: NORM/MAX/RAW
            self.writeCommand(':WAV:POIN:MODE %s'% mode)
        else:
            ret = self.writeCommand(':WAV:DATA? CHAN%d'% channel_idx, need_response=True)
     



# coding:utf8
__doc__ = 'DG811/812/821/822/831/832 Programmable Signal Generator'
__author__ = 'Peng Shulin <trees_peng@163.com>'
__license__ = 'MCUSH designed by Peng Shulin, all rights reserved.'
from re import compile as re_compile
from .. import Instrument
from .. import Env


class DG800( Instrument.SerialInstrument ):
    DEFAULT_NAME = 'DG800'
    DEFAULT_IDN = re_compile( '.*' )

    def __init__( self, *args, **kwargs ):
        Instrument.SerialInstrument.__init__( self, *args, **kwargs ) 
        raise NotImplementedError

        # TODO: append the instruments

